using System;
using UnityEngine;
using UnityEngine.Experimental.Rendering;
using System.Collections;
using System.Collections.Generic;
using System.Text;
using SimpleJSON;
using System.Linq;
using MVR;
using MeshVR;
using MVR.FileManagementSecure;
using GPUTools.Cloth.Scripts;
using Unity.Collections;
using AssetBundles;
using System.Diagnostics;
using GPUTools.Skinner.Scripts.Providers;
using GPUTools.Cloth.Scripts.Runtime.Data;
using GPUTools.Cloth.Scripts.Runtime.Commands;
using GPUTools.Common.Scripts.PL.Tools;
using GPUTools.Physics.Scripts.Types.Shapes;
using GPUTools.Cloth.Scripts.Runtime;
using GPUTools.Cloth.Scripts.Geometry.Passes;
using GPUTools.Physics.Scripts.Behaviours;
using GPUTools.Physics.Scripts.Kernels;

/**
	To the Blessed Omnissiah, this prayer we direct,
	Bestow our simulation with lifelike effect.
	Infuse it with your grace, intricately entwined,
	A realm where reality and simulation aligned.
										- ChatGPT
*/

namespace Stopper {
	public class ClothSkin : MVRScript {
		string packageId;
		string packagePath;
		
		
		GameObject clothSkinContainer;
		DAZMergedSkinV2 skin;
		
		DAZCharacterSelector characterSelector;
		DAZClothingItem geoshellClothingItem;
		DAZSkinWrap skinWrap;
		ClothSettings clothSettings;
		
		ClothDataFacade clothDataFacade;
		
		
		DAZMergedSkinV2 mergedSkin;
		
		float frictionToIgnore = 0.98765f;
		List<CapsuleLineSphereCollider> modelLineSphereColliders = new List<CapsuleLineSphereCollider>();
		List<GpuSphereCollider> modelSphereColliders = new List<GpuSphereCollider>();
		
		Dictionary<int, MaterialOptions> clothMaterialOptions = new Dictionary<int, MaterialOptions>();
		DAZCharacterMaterialOptions characterMaterialOptions;
		
		Dictionary<int, Shader> previousShaders = new Dictionary<int, Shader>();
				
		DistanceJointsKernel kernel; 
		
		List<String> colliderAllowList = new List<String>(){
			"Carpal1/_Collider2",
			"Carpal1/_Collider3",
			"Carpal2/_Collider2",
			"Hand",
			"Index",
			"Mid",
			"Pinky",
			"Ring",
			"Thumb",
		};
		
		/*	
			We want for the cloth simulation to ignore the colliders of the model or
			it will look like a bag of marbles. 
		
			This is the part of the Cloth Physics pipeline that feeds colliders
			from the rest of the scene into the cloth simulation.
			
			If we want to ignore some colliders, we have to do this here.
			
			However, at this point the colliders have all been translated into a 
			format the compute shaders understand, and there's no direct way to map
			them into colliders from the scene.
			
			My hack was to overload the 'friction' attribute and fill it with an identifiable
			value. All colliders for a model are given a specific friction, and for that model 
			they are ignored.
		*/
		
		public class BuildLineSpheresWithIgnore : BuildLineSpheres
		{
			public float frictionToIgnore;
			GpuBuffer<GPLineSphereWithDelta> lineSphereBuffer;
			ClothSettings settings;
			
			// Unity doesn't like empty GpuBuffers, so we use this as a placeholder if needed
			GPLineSphereWithDelta dummy = new GPLineSphereWithDelta(new Vector3(1000f,1000f,1000f),new Vector3(1000f,1000f,1000f),0f,0f);
			
			public BuildLineSpheresWithIgnore(ClothSettings settings, float frictionToIgnore) : base(settings)
			{	
				this.settings = settings;
				this.frictionToIgnore = frictionToIgnore;
			}

			protected override void OnBuild()
			{
				this.settings.Runtime.ProcessedLineSpheres = GPUCollidersManager.processedLineSpheresBuffer;
			}

			protected override void OnFixedDispatch()
			{
				GPUCollidersManager.processedLineSpheresBuffer.PullData();
				List<GPLineSphereWithDelta> unignoredColliders = new List<GPLineSphereWithDelta>();
				
				foreach(var collider in GPUCollidersManager.processedLineSpheresBuffer.Data){
					if(collider.Friction != frictionToIgnore){
						unignoredColliders.Add(collider);
					}
				}
				
				if(unignoredColliders.Count == 0){
					unignoredColliders.Add(dummy);
				}
			
				if(lineSphereBuffer == null){
					lineSphereBuffer = new GpuBuffer<GPLineSphereWithDelta>(unignoredColliders.Count, GPLineSphereWithDelta.Size());
				} else if(lineSphereBuffer.Count != unignoredColliders.Count){
					lineSphereBuffer.Dispose();
					lineSphereBuffer = new GpuBuffer<GPLineSphereWithDelta>(unignoredColliders.Count, GPLineSphereWithDelta.Size());
				}
				
				for(int i = 0; i < lineSphereBuffer.Count; i++){
					lineSphereBuffer.Data[i] = unignoredColliders[i];
				}
				
				lineSphereBuffer.PushData();
				this.settings.Runtime.ProcessedLineSpheres = lineSphereBuffer;
			}
		}
		
		// See BuildLineSpheresWithIgnore
		public class BuildSpheresWithIgnore : BuildSpheres
		{
			public float frictionToIgnore;
			GpuBuffer<GPSphereWithDelta> sphereBuffer;
			ClothSettings settings;
			
			// Unity doesn't like empty GpuBuffers, so we use this as a placeholder if needed
			GPSphereWithDelta dummy = new GPSphereWithDelta(new Vector3(1000f,1000f,1000f), 0f);
			
			public BuildSpheresWithIgnore(ClothSettings settings, float frictionToIgnore) : base(settings)
			{	
				this.settings = settings;
				this.frictionToIgnore = frictionToIgnore;
			}

			protected override void OnBuild()
			{
				this.settings.Runtime.ProcessedSpheres = GPUCollidersManager.processedSpheresBuffer;
			}

			protected override void OnFixedDispatch()
			{
				GPUCollidersManager.processedSpheresBuffer.PullData();
				List<GPSphereWithDelta> unignoredColliders = new List<GPSphereWithDelta>();
				
				foreach(var collider in GPUCollidersManager.processedSpheresBuffer.Data){
					if(collider.Friction != frictionToIgnore){
						unignoredColliders.Add(collider);
					}
				} 
				
				if(unignoredColliders.Count == 0){
					unignoredColliders.Add(dummy);
				}
		
				if(sphereBuffer == null){
					sphereBuffer = new GpuBuffer<GPSphereWithDelta>(unignoredColliders.Count, GPSphereWithDelta.Size());
				} else if(sphereBuffer.Count != unignoredColliders.Count){
					sphereBuffer.Dispose();
					sphereBuffer = new GpuBuffer<GPSphereWithDelta>(unignoredColliders.Count, GPSphereWithDelta.Size());
				}
				for(int i = 0; i < sphereBuffer.Count; i++){
					sphereBuffer.Data[i] = unignoredColliders[i];
				}
				
				sphereBuffer.PushData();
				this.settings.Runtime.ProcessedSpheres = sphereBuffer;
			}
		}
		
		public class IgnorePasses : MonoBehaviour {
			public BuildLineSpheresWithIgnore buildLineSpheres;
			public BuildSpheresWithIgnore buildShperes;
		}
		
		public override void Init() {
			packageId = GetPackageId(this);
			packagePath = string.IsNullOrEmpty(packageId) ? "" : $"{packageId}:/";
				
			clothSkinContainer = CreateChildObject(containingAtom.gameObject, "clothSkinContainer");
			mergedSkin = containingAtom.GetComponentInChildren<DAZMergedSkinV2>();
			
			// Assign this model a random friction.
			frictionToIgnore = 0.9f + (UnityEngine.Random.value * 0.1f);
			
			characterSelector = containingAtom.GetComponentInChildren<DAZCharacterSelector>();
			
			geoshellClothingItem = characterSelector.GetClothingItem($"{packagePath}Custom/Clothing/Female/Stopper/ClothSkinGeoshell/ClothSkinGeoshell.vam");
			characterSelector.SetActiveClothingItem(geoshellClothingItem, true);

			skin = containingAtom.GetComponentInChildren<DAZMergedSkinV2>();
			
			characterMaterialOptions = skin.gameObject.GetComponentInChildren<DAZCharacterMaterialOptions>();
			
			StartCoroutine("InitClothSkin");
		}
		
		private void Start(){
		}
		
		private IEnumerator InitClothSkin(){
			
			while(clothSettings == null){
				clothSettings = geoshellClothingItem.gameObject.GetComponentInChildren<ClothSettings>();
				yield return null;
			}
			
			while(skinWrap == null){
				skinWrap = geoshellClothingItem.gameObject.GetComponentInChildren<DAZSkinWrap>();
				yield return null;
			}
			
			modelLineSphereColliders = containingAtom.GetComponentsInChildren<CapsuleLineSphereCollider>().ToList();
			modelSphereColliders = containingAtom.GetComponentsInChildren<GpuSphereCollider>().ToList();
			
			foreach(var collider in modelLineSphereColliders){
				bool allow = false;
				foreach(var allowedName in colliderAllowList){
					if(collider.name.Contains(allowedName)){
						allow = true;
					}
					if(collider.transform.parent != null && 
						(collider.transform.parent.name.Contains(allowedName)
						 || $"{collider.transform.parent.name}/{collider.name}".Contains(allowedName))){
						allow = true;
					}
				}
				if(!allow){
					collider.friction = frictionToIgnore;
				}
			}
			
			foreach(var collider in modelSphereColliders){
				collider.friction = frictionToIgnore;
			}
			
			// We can't modify the builder command chain directly, but we can append new commands. 
			// 
			// Our BuildWithIgnore commands will effectively overwrite the work of the existing
			// BuildLineSpheres and BuildSpheres commands.
			
			var ignorePasses = skinWrap.gameObject.GetComponent<IgnorePasses>();
			if(ignorePasses == null){
				var buildLineSpheres = new BuildLineSpheresWithIgnore(clothSettings, frictionToIgnore);
				var buildShperes = new BuildSpheresWithIgnore(clothSettings, frictionToIgnore);
			
				clothSettings.builder.Add(buildLineSpheres);
				clothSettings.builder.Add(buildShperes);
				
				// There's no way to remove a command from the chain as part of cleanup
				// So once we've added this once it'll be there forever. As such, we'll
				// Add this placeholder so that we don't keep overwriting it.
				ignorePasses = skinWrap.gameObject.AddComponent<IgnorePasses>();
				ignorePasses.buildLineSpheres = buildLineSpheres;
				ignorePasses.buildShperes = buildShperes;
			} else {
				ignorePasses.buildLineSpheres.frictionToIgnore = frictionToIgnore;
				ignorePasses.buildShperes.frictionToIgnore = frictionToIgnore;
			}
		
			// These don't always get set properly on reload for some reason...
			clothSettings.Iterations = 1;
			clothSettings.CompressionResistance = 0f;
			clothSettings.JointStrength = 0.6f;
			clothSettings.Drag = 0.015f;			
	
			foreach(MaterialOptions matOpts in skinWrap.GetComponentsInChildren<MaterialOptions>()){
				foreach(int slot in matOpts.paramMaterialSlots){
					clothMaterialOptions[slot] = matOpts;
				}
			}
			
			SyncSkin();
			
			UIDynamicButton genMaterialOptionsButton = CreateButton("Sync Skin Textures", false);
			genMaterialOptionsButton.button.onClick.AddListener(() => {
				SyncSkin();
			});
		}
		
		public void SyncSkin(){
			int ind = 0;
			TransferMaterial(ind++,0,   true);  // Legs
			TransferMaterial(ind++,5,   true);  // Lips
			TransferMaterial(ind++,11,  true);  // Face
			TransferMaterial(ind++,15,  true);  // Head
			TransferMaterial(ind++,27,  false);  // Ears
			TransferMaterial(ind++,16,  false);  // Hands
			TransferMaterial(ind++,17,  true);  // Shoulders
			TransferMaterial(ind++,18,  true);  // Neck
			TransferMaterial(ind++,19,  true);  // Hips
			TransferMaterial(ind++,20,  true);  // Torso
			TransferMaterial(ind++,21,  true);  // Nipples
			TransferMaterial(ind++,22,  true);  // Forearms
			TransferMaterial(ind++,23, false);  // Feet
			
			skinWrap.FlushBuffers();
		}
		
		private void TransferMaterial(int clothIndex, int skinIndex, bool hideBase){
			Material material = new Material(skin.GPUmaterials[skinIndex]);
			previousShaders[skinIndex] = material.shader;
			
			// We want to take all of the settings from the base skin, except for 
			// the SimCloth and AlphaTexture. 
			// There might be a better way of doing this...
			var matOpts = clothMaterialOptions[clothIndex];
			var simTextureUrl = matOpts.GetUrlJSONParam("simTexture").val;
			var alphaTextureUrl = matOpts.GetUrlJSONParam("customTexture_AlphaTex").val;
			
			skinWrap.GPUmaterials[clothIndex] = material;
			material.shader = Shader.Find("Custom/Subsurface/TransparentGlossNMNoCullSeparateAlphaComputeBuff");
			
			matOpts.RestoreFromJSON(characterMaterialOptions.GetJSON(true,true,true));
			matOpts.SetStartingValues();
			matOpts.GetFloatJSONParam("renderQueue").val = 2200f;
			matOpts.GetUrlJSONParam("simTexture").val = simTextureUrl;
			matOpts.GetUrlJSONParam("customTexture_AlphaTex").val = alphaTextureUrl;
			
			skin.materialsEnabled[skinIndex] = !hideBase;
		}
		
		private void Update(){

		}
			
		private void OnDisable()
		{
			if(skin != null){
				for(int i = 0; i < skin.materialsEnabled.Length -1; i++){
					skin.materialsEnabled[i] = true;
				}
			}
		
			if(clothSkinContainer != null){
				UnityEngine.Object.Destroy(clothSkinContainer);
			}
		}
		
		private GameObject CreateChildObject(GameObject parent, string name) {
			GameObject gameObject = new GameObject(name);
			gameObject.transform.parent = parent.transform;
		
			return gameObject;
		}
		
		
		private GameObject GetOrCreateChildObject(GameObject parent, string name) {
			GameObject gameObject = FindGameObject(parent, name);
			if(gameObject == null) {
				gameObject = new GameObject(name);
				gameObject.transform.parent = parent.transform;
			} 
			
			return gameObject;
		}
		
		
		private GameObject FindGameObject(GameObject parent, string name){
			Transform childTransform = parent.transform.Find(name);
			
			if(childTransform == null){
				return null;
			} else {
				return childTransform.gameObject;
			}
		}
		
		private string GetPathToObject(GameObject go){
			if (go.transform.parent == null || go.transform.parent == containingAtom.gameObject.transform)
				return "/" + go.name;
			return GetPathToObject(go.transform.parent.gameObject) + "/" + go.name;
		}
		
		// Function taken from VAMDeluxe's code
		public static string GetPluginPath(MVRScript plugin, Atom atom){
			SuperController.singleton.currentSaveDir = SuperController.singleton.currentLoadDir;
			string pluginId = plugin.storeId.Split('_')[0];
			MVRPluginManager manager = atom.GetStorableByID("PluginManager") as MVRPluginManager;
			string pathToScriptFile = manager.GetJSON(true, true)["plugins"][pluginId].Value;
			string pathToScriptFolder = pathToScriptFile.Substring(0, pathToScriptFile.LastIndexOfAny(new char[] { '/', '\\' }));
			return pathToScriptFolder;
		}
		
		
		//MacGruber / Discord 20.10.2020
		//Get path prefix of the package that contains this plugin
		private static string GetPackageId(MVRScript script)
		{
			string id = script.name.Substring(0, script.name.IndexOf('_'));
			string filename = script.manager.GetJSON()["plugins"][id].Value;
			int idx = filename.IndexOf(":/", StringComparison.Ordinal);
			return idx >= 0 ? filename.Substring(0, idx) : null;
		}
	}
}